#!/usr/bin/env python

from multiprocessing import Pool
import os
import subprocess
import pyfits
import sys
import pickle
import argparse
import numpy as num

import pyLikelihood as pyLike
from gt_apps import GtApp

""" The overall structure and many of the underlying algorithms used
here are based on and sometimes blantenly copied from batch_tssmap.py
which was written by J. Chiang """

def resolve_fits_files(infile):

    '''This function returns a properly formatted infile loaation since
    most of the analysis is done in subdirectories and the actual support
    files are in ../.'''

    infile = '../'+infile
    foo = infile.strip('@')
    try:
        pyfits.open(foo)
    except IOError:
        return [item.strip() for item in open(foo)]
    return [infile]

def getPointSource(like):

    '''This function creats a test source object for use in the likelihood
    analysis at a specific pixel.'''

    try:
        test_src = pyLike.PointSource(0, 0, like.observation.observation)
    except AttributeError:
        test_src = pyLike.PointSource(0, 0, like.binnedData.observation)
    pl = pyLike.SourceFactory_funcFactory().create('PowerLaw')
    pl.setParamValues((1, -2, 100))
    indexPar = pl.getParam('Index')
    indexPar.setBounds(-3.5, -1)
    pl.setParam(indexPar)
    prefactor = pl.getParam('Prefactor')
    prefactor.setBounds(1e-10, 1e3)
    prefactor.setScale(1e-9)
    pl.setParam(prefactor)
    test_src.setSpectrum(pl)
    return test_src

def runLikelihood(subdir, tpl_file):

    '''This runction runs the likelihood code on a set of pixels in a
    subdirectory.  It takes as input the subdirectory to work on and a
    template counts map.  It reads it's configuration from a pickle
    file (par.pck) that should be located in the subdirectory and the
    pixel locations from another pickle file (pixel.pck).  It then
    creats an overall likelihood object, does a quick global fit and
    then loops over the pixels.  At each pixel, it creats a test
    source, fits that source, calculates the TS of the source and
    writes the results to an output file in the subdirectory called
    ts_results.dat.'''


    parfile = open("par.pck", "r")
    pars = pickle.load(parfile)

    pixelfile = open("pixel.pck", "r")
    pixels = pickle.load(pixelfile)

    pixel_coords = PixelCoords(tpl_file)

    if pars['statistic'] == 'UNBINNED':
        import UnbinnedAnalysis as UBAn
        obs = UBAn.UnbinnedObs(resolve_fits_files(pars['evfile']),
                               resolve_fits_files(pars['scfile']),
                               expMap='../'+pars['expmap'],
                               expCube='../'+pars['expcube'],
                               irfs=pars['irfs'])
        like = UBAn.UnbinnedAnalysis(obs, '../'+pars['srcmdl'], pars['optimizer'])
    elif pars['statistic'] == 'BINNED':
        import BinnedAnalysis as BAn 
        obs = BAn.BinnedObs(srcMaps='../'+pars['srcmaps'],
                            expCube='../'+pars['expcube'],
                            binnedExpMap='../'+pars['bexpmap'],
                            irfs=pars['irfs'])
        like = BAn.BinnedAnalysis(obs, '../'+pars['srcmdl'], pars['optimizer'])

    like.setFitTolType(pars['toltype'])
    like.optimize(0)
    loglike0 = like()
    test_src = getPointSource(like)
    target_name = 'testSource'
    test_src.setName(target_name)
    outfile = 'ts_results.dat'
    finished_pixels = []
    if os.path.isfile(outfile):
        input = open(outfile, 'r')
        for line in input:
            tokens = line.strip().split()
            ij = int(tokens[0]), int(tokens[1])
            finished_pixels.append(ij)
        input.close()
    output = open(outfile, 'a')
    for indx, i, j in pixels:
        if (i, j) in finished_pixels:
            continue
        ra, dec = pixel_coords(i, j)
        test_src.setDir(ra, dec, True, False)
        like.addSource(test_src)
        like.optimize(0)
        ts = -2*(like() - loglike0)
        output.write("%3i  %3i %.3f  %.3f  %.5f\n" % (i, j, ra, dec, ts))
        output.flush()
        like.deleteSource(target_name)
    output.close()

def launchJobs(SQ):

    '''This function actually launches a job.  It takes the configuration
    square as input, creates the subdirectories, writes the par and
    pixel pickle files and then launches the code again with the
    subdirectory and template file as options with the number of jobs
    requested.'''

    subdir = SQ[0]
    pars = SQ[1]
    pixels = SQ[2]
    tpl_file = SQ[3]

    os.chdir(subdir)
    
    parfile = open("par.pck", "w")
    pickle.dump(pars, parfile)
    parfile.close()

    pixelfile = open("pixel.pck", "w")
    pickle.dump(pixels, pixelfile)
    pixelfile.close()

    if sys.argv[0][0] == '/':
        this_arg = sys.argv[0]
    else:
        this_arg = '../'+sys.argv[0]
    
    command = ("python %s runLike %s %s"
               % (this_arg,subdir, tpl_file))
    sys.stdout.write("partition %s: " % subdir)
    print command
    sys.stdout.flush()
    subprocess.call(command.split())
    
def _readlines(fileobj, comment=''):

    '''This function read a line in the saved data in the
    subdirectories.'''

    lines = []
    for line in fileobj:
        if ((comment != '' and line.find(comment) == 0)
            or len(line.strip()) == 0):
            continue
        if comment != '':
            line = line.split(comment)[0].strip()
        lines.append(line)
    return lines

def read_data(file, delimiter=None, nskip=0, ncols=0, nmax=0, comment="#"):

    '''This function reads in the saved data in the subdirectories.'''

    data = _readlines(open(file), comment=comment)
    if nmax == 0:
        nmax = len(data) - nskip
    data = data[nskip:nskip+nmax]
    if ncols == 0:
        ncols = len(data[0].split(delimiter))
    columns = []
    for i in range(ncols):
        columns.append([])
    for line in data:
        datum = line.split(delimiter)
        for i in range(ncols):
            if (datum[i].find('.') == -1 and datum[i].find('e') == -1
                and datum[i].find('E') == -1):
#                columns[i].append(int(datum[i]))
                columns[i].append(float(datum[i]))
            else:
                columns[i].append(float(datum[i]))
    for i in range(ncols):
        columns[i] = num.array(columns[i])
    return tuple(columns)

class PixelCoords():

    '''This class retuns the pixel coordinates of a fits file.'''

    def __init__(self, template_file):
        self.proj = pyLike.SkyProj(template_file)
        template = pyfits.open(template_file)
        self.nx = template[0].header['NAXIS1']
        self.ny = template[0].header['NAXIS2']
    def __call__(self, i, j):
        return self.proj.pix2sph(i+1, j+1)

class BatchTsMap(object):

    '''Generates a TS map on a multicore machine.'''


    def __init__(self, pars, num_queues=40, tpl_file='cmap_tpl.fits',savetmp=True):

        '''Initializes the object and sets up the number of queues and stores
        all the parameters.'''

        self.num_queues = num_queues
        self.tpl_file = tpl_file
        self.savetmp = savetmp
        self.pars = pars

        self._createTemplate()
        self.pixel_coords = PixelCoords(self.tpl_file)
        self.njobs = min(self.num_queues, self.pixel_coords.ny)
        self.subdirs = [("%03i" % i) for i in range(self.njobs)]
        
    def _createTemplate(self):
        pars = self.pars
        gtbin = GtApp('gtbin')
        gtbin.run(algorithm='CMAP', evfile=pars['evfile'],
                  scfile=pars['scfile'], outfile=self.tpl_file,
                  nxpix=pars['nxpix'], nypix=pars['nypix'],
                  binsz=pars['binsz'], coordsys=pars['coordsys'],
                  xref=pars['xref'], yref=pars['yref'], axisrot=0,
                  proj=pars['proj'], chatter=0, clobber='yes')

    def getPixels(self, partition):
        nx, ny = self.pixel_coords.nx, self.pixel_coords.ny
        nsize = nx*ny/self.njobs
        if nx*ny % self.njobs != 0:
            nsize += 1
        pixels = []
        indx = 0
        for i in range(nx):
            for j in range(ny):
                if indx >= nsize*partition and indx < nsize*(partition + 1):
                    pixels.append((indx, i, j))
                indx += 1
        return pixels

    def prepareSubdirs(self):
        for subdir in self.subdirs:
            try:
                os.mkdir(subdir)
            except OSError:
                pass

    def merge_results(self):
        tsmap = pyfits.open(self.tpl_file)
        tsmap[0].data = num.zeros(tsmap[0].data.shape, dtype=num.float)
        for subdir in self.subdirs:
            result_file = os.path.join(subdir, 'ts_results.dat')
            try:
                ii, jj, ra, dec, ts_vals = read_data(result_file)
            except:
                continue
            for i, j, ts in zip(ii, jj, ts_vals):
                tsmap[0].data[j][i] = ts
        tsmap.writeto(self.pars['outfile'], clobber=True)
                
    def remove_tempfiles(self):

        for subdir in self.subdirs:
            os.remove(subdir+"/par.pck")
            os.remove(subdir+"/pixel.pck")
            os.remove(subdir+"/ts_results.dat")
            os.removedirs(subdir)
        os.remove(self.tpl_file)


def gttsmap_mp(pars,num_queues,savetmp):

    tsmap = BatchTsMap(pars,num_queues,savetmp=savetmp)
    tsmap.prepareSubdirs()
    
    SQ = [(subdir,tsmap.pars,tsmap.getPixels(int(subdir)),tsmap.tpl_file) for subdir in tsmap.subdirs]
    pool = Pool(processes=tsmap.njobs)
    pool.map(launchJobs,SQ)
    
    tsmap.merge_results()
    if not savetmp:
        print "Deleting temporary files."
        tsmap.remove_tempfiles()


def cli():

    helpString = "Generates a TS Map by running seperate pixels on seperate threads.\
                  It creates a template counts map to determine the location of the \
                  pixels and the calculates the TS of a test source at each pixel on \
                  that map.  It then merges the results at the end.  The map is divided\
                  into seperate regions (called 000,001,002...) depending on the number\
                  of jobs the user requests.  It creates subdirectories for each region\
                  and operates on them within the subdirectory.  For more details on\
                  the parameters see the gttsmap help file.  NOTE:  ONLY DOES AN \
                  UNBINNED ANALYSIS.  NOTE2:  Any files referenced in the model XML file\
                  must be referenced by abosolute and not relative directories \
                  (ie. not ./ or ../) \
                  \
                  BEWARE:  THIS PROGRAM IS KNOWN TO PRODUCE SPURIOUS RESULTS IN SOME CASES.\
                                             USE WITH CAUTION."

    parser = argparse.ArgumentParser(description=helpString)

    parser.add_argument("nxpix", type=int, help="Number of pixels along x-axis.  See gttsmap help for more information.")
    parser.add_argument("nypix", type=int, help="Number of pixels along y-axis.  See gttsmap help for more information.")
    parser.add_argument("jobs", type=int, help="The number of concurrent jobs.")
    parser.add_argument("evfile", help="Input event file.  See gttsmap help for more information.")
    parser.add_argument("scfile", help="The spacecraft data file. See gttsmap help for more information.")    
    parser.add_argument("expcube", help="Input livetime cube.  See gttsmap help for more information.")
    parser.add_argument("srcmdl", help="XML source model definition.  Any files in the xml (like the diffuse models) need to be referenced by absolute directories. See gttsmap help for more information.")
    parser.add_argument("IRFS", help="IRFs to use.  See gttsmap help for more information.")
    parser.add_argument("optimizer", help="The optimizer (e.g. NEWMINUIT). See gttsmap help for more information.")
    parser.add_argument("ftol", type=float, help="Fit tolerance. See gttsmap help for more information.")
    parser.add_argument("toltype", type=int, help="Tolerance type (0 for RELATIVE and 1 for ABSOLUTE). See gttsmap help for more information.")
    parser.add_argument("binsz", type=float, help="Image scale (deg/pix).  See gttsmap help for more information.")
    parser.add_argument("coordsys", help="CEL or GAL.  See gttsmap help for more information.")
    parser.add_argument("xref", type=float, help="x-coord of center (RA or l).  See gttsmap help for more information.")
    parser.add_argument("yref", type=float, help="y-coord of center (DEC or b).  See gttsmap help for more information.")
    parser.add_argument("proj", help="Coordinate projection. See gttsmap help for more information.")
    parser.add_argument("outfile", help="Output file name.")

    subparsers = parser.add_subparsers(help="UNBINNED or BINNED. See gttsmap help for more information.", dest="statistic")

    unbinned_parser = subparsers.add_parser('UNBINNED')
    unbinned_parser.add_argument("expmap", help="Input exposure map.  See gttsmap help for more information.")    

    binned_parser = subparsers.add_parser('BINNED')
    binned_parser.add_argument("cmap", help="Counts map file.  See gttsmap help for more information.")
    binned_parser.add_argument("bexpmap", help="Binned exposure map.  See gttsmap help for more information.")
    binned_parser.add_argument("srcmaps", help="Source maps file.  See gtlike help for more information.")

    parser.add_argument("--savetmp", default = False, help="Save the temporary files (default is False).")
    
    args = parser.parse_args()

    pars = {}

    pars['statistic'] = args.statistic
    pars['nxpix'] = args.nxpix
    pars['nypix'] = args.nypix
    pars['binsz'] = args.binsz
    pars['coordsys'] = args.coordsys
    pars['xref'] = args.xref
    pars['yref'] = args.yref
    pars['proj'] = args.proj
    pars['expcube'] = args.expcube
    pars['srcmdl'] = args.srcmdl
    pars['outfile'] = args.outfile
    pars['irfs'] = args.IRFS
    pars['optimizer'] = args.optimizer
    pars['ftol'] = args.ftol
    pars['toltype'] = args.toltype
    pars['evfile'] = args.evfile
    pars['scfile'] = args.scfile

    if(args.statistic == 'UNBINNED'):
        pars['expmap'] = args.expmap
    elif(args.statistic == 'BINNED'):
        pars['cmap'] = args.cmap
        pars['bexpmap'] = args.bexpmap
        pars['srcmaps'] = args.srcmaps

    gttsmap_mp(pars,num_queues=args.jobs,savetmp=args.savetmp)

def cli_runLike():

    helpString = "This is a helper function that runs each individual partition.\
                  Don't use it manually."

    parser = argparse.ArgumentParser(description=helpString)
    parser.add_argument("runLike", help="Flag to indicate that you are running on a partition.")
    parser.add_argument("subdir", help="Subdirectory to run under.")
    parser.add_argument("tpl_file", help="File name of the template CMAP.")

    args = parser.parse_args()

    runLikelihood(args.subdir, "../"+args.tpl_file)



if __name__ == '__main__':

    if len(sys.argv) > 1:
        if sys.argv[1] == 'runLike':
            cli_runLike()
        else:
            cli()
    else:
        cli()




#!/usr/bin/env python

from multiprocessing import Pool
import numpy as np
import tempfile
import os
import subprocess
from gtapps_mp.utils import pyfits

from gt_apps import expMap

def expmap(square):

    '''This is the atomic function that actually runs in the seperate
    threads.  The input is a python list.  The first element is a list
    of two arrays giving the x and y points.  The second through last
    elements are the scfile, evfile, expcube, nlong, nlat, irfs,
    srcrad and nenergies.  This function creates a temporary file for
    the output and returns that file's name.'''

    print "Starting calculation of region {},{} to {},{}".format(square[0][0][0],
                                                                 square[0][1][0],
                                                                 square[0][0][1],
                                                                 square[0][1][1])
    osfilehandle,outfilename = tempfile.mkstemp(suffix=".fits")
    expMap['nlongmin'] = "{:,.0f}".format(float(square[0][0][0]))
    expMap['nlongmax'] = "{:,.0f}".format(float(square[0][0][1]))
    expMap['nlatmin'] = "{:,.0f}".format(float(square[0][1][0]))
    expMap['nlatmax'] = "{:,.0f}".format(float(square[0][1][1]))
    expMap['scfile'] = square[1]
    expMap['evfile'] = square[2]
    expMap['expcube'] = square[3]
    expMap['nlong'] = square[4]
    expMap['nlat'] = square[5]
    expMap['irfs'] = square[6]
    expMap['srcrad'] = square[7]
    expMap['nenergies'] = square[8]
    expMap['outfile'] = outfilename
    expMap['submap'] = "yes"
    expMap['chatter'] = 0
    expMap.run(print_command=False)
    print "Completed calculation of region {},{} to {},{}".format(square[0][0][0],
                                                                  square[0][1][0],
                                                                  square[0][0][1],
                                                                  square[0][1][1])
                
    return outfilename

def expsum(filenames, Outfile, SaveTemp):

    '''This function takes a list of exposure maps and adds them together.
    If there is only one file to be added, it just copies it to the
    outfile.  If there is more than one, it uses pyfits to open them all
    up, sum the first hdus (the data) and then replaces the first file's
    primary hdu with this summed data and writes it to the output file.'''

    if len(filenames) <= 1:
        subprocess.call(["cp", filenames[0], Outfile])
    else:
        expmap_files = [pyfits.open(filename) for filename in filenames]
        summed_expmap_hdu = (np.array([expmap_file[0].data for expmap_file in expmap_files])).sum(axis=0)
        expmap_files[0][0].data = summed_expmap_hdu
        expmap_files[0][0].update_header()
        expmap_files[0].writeto(Outfile, clobber='True')
        for expmap_file in expmap_files: expmap_file.close()
        
    if SaveTemp:
        print "Did not delete the following temporary files:"
        print filenames
    else:
        print "Deleting temporary files..."
        for filename in filenames:
            os.remove(filename)

def gtexpmap_mp(nlong, nlat, xbins, ybins, SCFile, EVFile, ExpCube, 
                IRF, srcrad, nenergy, OutFile, SaveTemp):

    '''This function determines the x,y pairs for dividing up the jobs on
    the sky.  It then creates a list of tuples conatining all of the
    information needed for the atomic expmap function.'''

    bins = xbins*ybins

    if np.mod(nlong,xbins) or np.mod(nlat,ybins):
        print "The number of x and y bins must fit evenly into the number of long or lat points."
        return

    stepx = nlong/xbins
    stepy = nlat/ybins

    x_start = np.linspace(0,nlong-stepx,xbins)
    x_stop = np.linspace(stepx,nlong,xbins)
    y_start = np.linspace(0,nlat-stepy,ybins)
    y_stop = np.linspace(stepy,nlat,ybins)

    pairs = [(x_pair,y_pair) for x_pair in np.column_stack((x_start,x_stop)) for y_pair in np.column_stack((y_start,y_stop))]
    SQ = [(row, SCFile,EVFile,ExpCube,nlong,nlat,IRF,srcrad,nenergy) for row in pairs]    

    pool = Pool(processes=bins)      
    print "Spawning {} jobs...".format(bins)
    filenames = pool.map(expmap,SQ)
    print "Combining temporary files..."
    expsum(filenames, OutFile, SaveTemp)

def cli():

    helpString = "Submits the gtexpmap program as sperate threads via python and\
                  joins up the resulting temporary exposure maps at the end\
                  resulting in a single exposure map for the input event file.\
                  This greatly reduces the running time. For more details on \
                  gtexpmap see the gtexpmap help file.  Note that the checksum \
                  and datasum are incorrect for the final file.  The number of \
                  spawned jobs is equal to xbins x ybins."

    import argparse

    parser = argparse.ArgumentParser(description=helpString)
    parser.add_argument("nlong", type=int, help="Number of longitude points.  See gtexpmap help for more information.")
    parser.add_argument("nlat", type=int, help="Number of latitude points.  See gtexpmap help for more information.")
    parser.add_argument("xbins", type=int, help="The number of bins along the x-axis.  Must divide evenly into nlong.")
    parser.add_argument("ybins", type=int, help="The number of bins along the y-axis.  Must divide evenly into nlat.")
    parser.add_argument("sfile", help="The spacecraft data file. See gtexpmap help for more information.")
    parser.add_argument("evfile", help="Input event file.  See gtexpmap help for more information.")
    parser.add_argument("expcube", help="Input livetime cube.  See gtexpmap help for more information.")
    parser.add_argument("IRFS", help="IRFs to use.  See gtexpmap help for more information.")
    parser.add_argument("srcrad", help="Radius.  See gtexpmap help for more information.")
    parser.add_argument("nenergies", help="Number of energy slices.  See gtexpmap help for more information.")
    parser.add_argument("outfile", help="Output file name.")

    parser.add_argument("--savetmp", default = False, help="Save the temporary files (default is False).")
    
    args = parser.parse_args()

    gtexpmap_mp(args.nlong, args.nlat, args.xbins, args.ybins, args.sfile, args.evfile, args.expcube, 
                args.IRFS, args.srcrad, args.nenergies, args.outfile, args.savetmp)


if __name__ == '__main__': cli()

    



#!/usr/bin/env python

from multiprocessing import Pool
import numpy as np
import tempfile
import os
import subprocess
from gtapps_mp.utils import pyfits

from gt_apps import evtbin,srcMaps

def run_gtsrcmaps(square):

    '''This is the atomic function that actually runs in the seperate
    threads.  The input is an array with two elements, the first describes
    the energy binning and the second lists all of the progromatic
    options.  This function runs evtbin to generate a counts cube for the
    sub-energy range and then gtsrcmaps.  It returns the bin number and
    the output filename of the srcmaps file for the sub-energy range.'''


    idx = square[0][0][0]
    emin = square[0][0][1]*0.001
    emax = square[0][-1][2]*0.001
    enumbins = len(square[0])
    options = square[1]

    ccubehandle,ccubefilename = tempfile.mkstemp(suffix=".fits")
    evtbin['evfile'] = options['evfile']
    evtbin['outfile'] = ccubefilename
    evtbin['algorithm'] = 'CCUBE'
    evtbin['nxpix'] = options['nxpix']
    evtbin['nypix'] = options['nypix']
    evtbin['binsz'] = options['binsz']
    evtbin['coordsys'] = options['coordsys']
    evtbin['xref'] = options['xref']
    evtbin['yref'] = options['yref']
    evtbin['axisrot'] = options['axisrot']
    evtbin['proj'] = options['proj']
    evtbin['ebinalg'] = options['ebinalg']
    evtbin['emin'] = emin
    evtbin['emax'] = emax
    evtbin['enumbins'] = enumbins
    evtbin['scfile'] = options['scfile']
    evtbin.run(print_command=False)

    srcmaphandle,srcmapfilename = tempfile.mkstemp(suffix=".fits")
    srcMaps['scfile'] = options['scfile']
    srcMaps['expcube'] = options['expcube']
    srcMaps['cmap'] = ccubefilename
    srcMaps['srcmdl'] = options['srcmdl']
    srcMaps['bexpmap'] = options['bexpmap']
    srcMaps['outfile'] = srcmapfilename
    srcMaps['irfs'] = options['irfs']
    srcMaps['chatter'] = 2
    srcMaps['emapbnds'] = "no"
    srcMaps.run(print_command=False)

    return idx,srcmapfilename

def srcmapssum(results, ref_hdu, outfile, savetemp):

    '''This function takes a list of source maps and joins them together.
    If there is only one file to be joined, it just copies it to the
    outfile.  If there is more than one, it uses pyfits to open them
    all up, add the energy slices together and then outputs this.  The
    first element in the list is the index.'''

    results.sort()

    if len(results) <= 1:
        subprocess.call(["cp", results[0][1], outfile])
    else:
        SrcMaps = [pyfits.open(result[1]) for result in results]
        for idx,sourceHDU in enumerate(SrcMaps[0][3:]):
            newHDU = SrcMaps[0][idx+3].copy()
            newHDU.data = np.delete(SrcMaps[0][idx+3].data,-1,0)
            for SrcMap in SrcMaps[1:]:
                newHDU.data = np.append(newHDU.data,SrcMap[idx+3].data[:-1],axis=0)
            newHDU.data = np.append(newHDU.data,[SrcMaps[-1][idx+3].data[-1]],axis=0)
            #newHDU.add_checksum()
            ref_hdu.append(newHDU)
        ref_hdu.writeto(outfile,clobber='yes')
        
    if savetemp:
        print "Did not delete the following temporary files:"
        print results
    else:
        print "Deleting temporary files..."
        for result in results:
            os.remove(result[1])

def gtsrcmaps_mp(nxpix, nypix, binsz, scfile, evfile, expcube, 
                 ccube, coordsys, proj, xref, yref, axisrot, ebinalg,
                 emin, emax, enumbins,jobs, srcmdl, bexpmap, irfs, 
                 outfile, savetmp):

    '''This function opens up the input counts map and divides up the
    energy bins into seperate jobs which are sent to the pool to be
    run.'''

    ref_hdu = pyfits.open(ccube)
    energies = np.array(ref_hdu[1].data)
    
    #There's an error when you only use one bin so you need to
    #make sure and compute more than one bin per job.
    while len(energies)/float(jobs) < 3:
        print "Too many jobs ({}), reducing by 1".format(jobs)
        jobs -= 1
        print "Jobs is now {}".format(jobs)
    energy_arrays = np.array_split(energies,jobs)

    options = {'nxpix': nxpix,
               'nypix': nypix,
               'binsz': binsz,
               'scfile': scfile,
               'evfile': evfile,
               'coordsys': coordsys,
               'xref': xref,
               'yref': yref,
               'axisrot': axisrot,
               'proj': proj,
               'ebinalg': ebinalg,
               'emin': emin,
               'emax': emax,
               'enumbins': enumbins,
               'expcube': expcube,
               'srcmdl': srcmdl,
               'bexpmap': bexpmap,
               'irfs': irfs}

    SQ = [(array,options) for array in energy_arrays]

    pool = Pool(processes=jobs)
    print "Spawning {} jobs...".format(jobs)
    results = pool.map(run_gtsrcmaps,SQ)
    print "Combining temporary files..."
    srcmapssum(results, ref_hdu, outfile, savetmp)

def cli():

    helpString = "Submits the gtsrcmaps program as sperate threads via python and\
                  joins up the resulting temporary source maps at the end\
                  resulting in a single source map file.\
                  This greatly reduces the running time. For more details on \
                  gtsrcmaps see the gtsrcmaps help file.  Note that the checksum \
                  and datasum are incorrect for the final file."

    import argparse

    parser = argparse.ArgumentParser(description=helpString)
    
    parser.add_argument('jobs', type=int, help="Number of concurrent jobs (threads) to run.")
    parser.add_argument('nxpix', type=int, help="Number of horizontal pixels.  See gtbin for more information.")
    parser.add_argument('nypix', type=int, help="Number of vertical pixels.  See gtbin for more information.")
    parser.add_argument('binsz', type=float, help="Degrees per pixel.  See gtbin for more information.")
    parser.add_argument('scfile', help="Spacecraft file.  See gtsrcmaps for more information.")
    parser.add_argument('evfile', help="Event file.  See gtbin for more information.")
    parser.add_argument('coordsys', help="Coordinate system (CEL or GAL).  See gtbin for more information.")
    parser.add_argument('xref', type=float, help="Horizontal position. See gtbin for more information.")
    parser.add_argument('yref', type=float, help="Vertical position. See gtbin for more information.")
    parser.add_argument('axisrot', type=float, help="Rotation angle.  See gtbin for more information.")
    parser.add_argument('proj', help="Coordintate projection. See gtbin for more information.")
    parser.add_argument('ebinalg', help="Energy bin specification.  See gtbin for more information.")
    parser.add_argument('emin', help="Minimum energy (in MeV). See gtbin for more information.")
    parser.add_argument('emax', help="Maximum energy (in MeV). See gtbin for more information.")
    parser.add_argument('enumbins', help="Number of energy bins. See gtbin for more information.")
    parser.add_argument('expcube', help="Output of gtltcube. See gtsrcmaps for more information.")
    parser.add_argument('srcmdl', help="XML model file.   See gtsrcmaps for more information.")
    parser.add_argument('bexpmap', help="Binned exposure map from gtexpcube2.   See gtsrcmaps for more information.")
    parser.add_argument('ccube', help="Counts cube produced by gtbin.    See gtsrcmaps for more information.")
    parser.add_argument('irfs', help="Name of the IRFs.   See gtsrcmaps for more information.")
    parser.add_argument('outfile', help="Name of the output FITS file")
    
    parser.add_argument("--savetmp", default = False, help="Save the temporary files (default is False).")
    
    args = parser.parse_args()

    gtsrcmaps_mp(args.nxpix, args.nypix, args.binsz, args.scfile, args.evfile, args.expcube, 
                 args.ccube, args.coordsys, args.proj, args.xref, args.yref, args.axisrot, args.ebinalg,
                 args.emin, args.emax, args.enumbins, args.jobs, args.srcmdl, args.bexpmap, args.irfs, 
                 args.outfile, args.savetmp)

if __name__ == '__main__': cli()

    



#!/usr/bin/env python

from multiprocessing import Pool
import numpy as np
from gtapps_mp.utils import pyfits
import tempfile
import os
import subprocess
import sys

from gt_apps import filter,expCube

def ltcube(times):

    '''This is the atomic function that actually runs in the seperate
    threads.  It takes a list as input where the first element is
    tmin, second is tmax, third is spacecraft file, fourth is the
    event file and fifth is the zmax parameter.  It first uses
    gtselect with wide open cuts to divide up the event file then it
    runs gtltcube on that event file.  The temporary event file is
    deleted automatically.  The function returns the name of the
    created ltcube file which can be combined with other files and/or
    deleted later.'''

    print "Starting calculation on interval {} to {}".format(times[0],times[1])
    if times[3] != '':
        evfile = tempfile.NamedTemporaryFile(suffix=".fits")
        filter['rad'] = "INDEF"
        filter['evclass'] = 0
        filter['infile'] = times[3]
        filter['outfile'] = evfile.name
        filter['ra'] = "INDEF"
        filter['dec'] = "INDEF"
        filter['tmin'] = times[0]
        filter['tmax'] = times[1]
        filter['emin'] = 0
        filter['emax'] = 400000
        filter['zmax'] = 180
        filter['convtype'] = -1
        filter['chatter'] = 0
        filter.run(print_command=True)

    osfilehandle,outfilename = tempfile.mkstemp(suffix=".fits")
    if times[3] != '':
        expCube['evfile'] = evfile.name
    else:
        expCube['evfile'] = ""
    expCube['scfile'] =  times[2]
    expCube['outfile'] = outfilename
    expCube['tmin'] = times[0]
    expCube['tmax'] = times[1]
    expCube['dcostheta'] = 0.025
    expCube['binsz'] = 1
    expCube['phibins'] = 0
    expCube['zmax'] = times[4]
    expCube['chatter'] = 0
    expCube.run(print_command=True)
    print "Completed calculation on interval {} to {}".format(times[0],times[1])
    return outfilename


def ltsum(filenames, Outfile, SaveTemp):

    '''This function takes a list of livetime cubes and sums them up using
    gtltsum.  It first checks to see if there's only one temporary
    file.  If so, it just copies that to the output file.  If not, it
    creates a temporary file that lists the individual ltcube files
    and operates gtltsum on them.'''

    if len(filenames) <= 1:
        subprocess.call(["cp", filenames[0], Outfile])
    else:
        fileListfile = tempfile.NamedTemporaryFile()
        for filename in filenames:
            fileListfile.file.write(filename + "\n")
        fileListfile.flush()
        subprocess.call(["gtltsum", 
                         "infile1=@"+fileListfile.name, 
                         "outfile="+Outfile])

    if SaveTemp:
        print "Did not delete the following temporary files:"
        print filenames
    else:
        print "Deleting temporary files..."
        for filename in filenames:
            os.remove(filename)


def gtltcube_mp(bins, SCFile, EVFile, OutFile, SaveTemp, zmax, tmin, tmax):

    '''This functions looks at a spacecraft file and splits the time into
    chunks that match the bin edges in the spacecraft file.  It then
    submits jobs based upon those start and stop times.  This is to
    make the resulting files as close to the original as possible.
    Note that this assumes you are using the full time period in your
    spacecraft file.'''

    verbose = False

    if EVFile != "":
        evfile = pyfits.open(EVFile, mode='readonly')
        gti_data = evfile[2].data

    if tmin == 0: 
        print "Determining start and stop times from the event file..."
        tstart = evfile[0].header['TSTART']
        tstop = evfile[0].header['TSTOP']
    else:
        print "Using user defined tmin and tmax..."
        tstart = tmin
        tstop = tmax

    print "Opening SC file to determine break points..."
    hdulist = pyfits.open(SCFile, mode='readonly')
    scdata = hdulist[1].data
    hdulist.close()
    scstart = scdata.field('START')
    scstop = scdata.field('STOP')

    time_filter = (tstart <= scstart) & (scstop <= tstop)

    redo = True
    if EVFile !="":
        print "Checking for good times in the event file..."
    while redo:
        redo = False
        scstartssplit = np.array_split(scstart[time_filter],int(bins))
        scstopssplit = np.array_split(scstop[time_filter],bins) 
    
        #Explicitly set the first and last point to the values in the evfile header
        scstartssplit[0][0] = tstart
        scstopssplit[-1][-1] = tstop

        starts = [st[0] for st in scstartssplit]
        stops = [st[-1] for st in scstopssplit]

        if EVFile != "":
            for interval in zip(starts,stops):
                if verbose: print "Looking at interval",interval[0],"to",interval[1]
                good_times = False
                #grrrr.  some bug in pyfits doesn't let me do this the python way...
                for gti_i in range(len(gti_data)):
                    if(not good_times):
                        if verbose: print "   Checking gti",gti_data[gti_i]['START'],"to",gti_data[gti_i]['STOP']
                        gti_starts = interval[0] <= gti_data[gti_i]['START'] <= interval[1]
                        gti_stops = interval[0] <= gti_data[gti_i]['STOP'] <= interval[1]
                        if verbose: print "   Does this gti start inside this interval? ", gti_starts
                        if verbose: print "   Does this gti stop inside this interval? ", gti_stops
                        good_times = gti_starts or gti_stops
                        if verbose: print
        
                if verbose: print "  Are there good times inside this interval? ", good_times
                if not good_times:
                    redo = True
                if verbose: print

        if redo:
            if bins <= 1:
                print "No good time intervals found.  Bailing..."
                sys.exit(1)
            print "One (or more) of the slices doesn't have a GTI."
            print "Reducing the number of threads from ",bins,"to",bins-1
            bins -= 1
        
    scfiles = [SCFile for st in scstartssplit]
    evfiles = [EVFile for st in scstartssplit]
    print "EVFiles:",evfiles
    zmaxes =  [zmax for st in scstartssplit]

    pool = Pool(processes=bins)      
    times = np.array([starts,stops,scfiles,evfiles,zmaxes])
    print "Spawning {} jobs...".format(bins)
    tempfilenames = pool.map(ltcube,times.transpose())
    print "Combining temporary files..."
    ltsum(tempfilenames, OutFile, SaveTemp)

def cli():

    helpString = "Submits the gtltcube program as sperate threads via python and\
                  joins up the resulting temporary exposure cubes at the end\
                  resulting in a single exposure cube for the input event file.\
                  This greatly reduces the running time. For more details on \
                  gtltcube see the gtltcube help file."

    import argparse

    parser = argparse.ArgumentParser(description=helpString)
    parser.add_argument("jobs", type=int, help="The number of jobs you wish to spawn (usually the number of cores on your machine).")
    parser.add_argument("sfile", help="The spacecraft data file. See gtltcube help for more information.")
    parser.add_argument("evfile", help="Input event file.  See gtltcube help for more information.")
    parser.add_argument("outfile", help="Output file name.")

    parser.add_argument("--savetmp", default = False, help="Save the temporary files (default is False).")
    parser.add_argument("--zmax", type=int, default = 180, help="zmax parameter for gtltcube (default is 180)")
    parser.add_argument("--tmin", type=float, default=0, help="start time (if not given, will derive from evfile)")
    parser.add_argument("--tmax", type=float, default=0, help="stop time (if not given, will derive from evfile)")
    
    args = parser.parse_args()

    gtltcube_mp(args.jobs, args.sfile, args.evfile, args.outfile, args.savetmp, args.zmax, args.tmin, args.tmax)


if __name__ == '__main__': cli()

    



#!/usr/bin/evn python

import sys
import os
from setuptools import setup, find_packages

setup(name='gtapps_mp',
      version='1.5',
      description='Fermi LAT Multicore Scripts',
      author='Jeremy S. Perkins (FSSC)',
      author_email='fermihelp@milkyway.gsfc.nasa.gov',
      url='http://fermi.gsfc.nasa.gov/ssc/',
      packages=find_packages(exclude='tests'),
      entry_points= {'console_scripts': [
                      'gtdiffrsp_mp = gtapps_mp.gtdiffrsp_mp:cli',
                      'gtexpmap_mp = gtapps_mp.gtexpmap_mp:cli',
                      'gtltcube_mp = gtapps_mp.gtltcube_mp:cli',
                      'gttsmap_mp = gtapps_mp.gttsmap_mp:cli',
                      ]},
      )



try:
    import astropy
except ImportError:
    has_astropy = False
else:
    has_astropy = True

if has_astropy:
    from astropy.io import fits as pyfits
    import astropy.wcs as pywcs
else:
    import pyfits
    import pywcs
                    



#!/usr/bin/env python

from multiprocessing import Pool
import numpy as np
from gtapps_mp.utils import pyfits
import tempfile
import os
import subprocess

from gt_apps import diffResps,filter

def diffrsp(times):

    '''This is the atomic function that actually runs in the seperate
    threads.  It takes a list as input where the first element is
    tmin, second is tmax, third is the spacecraft file, fourth is the
    event file, fifth is the source model and sixth is the IRF.  It
    first uses gtselect with wide open cuts to divide up the event
    file then it runs gtdiffrsp on that event file.  The temporary
    event file is deleted automatically.  The function returns the
    name of the created event file which can be combined with other
    files and/or deleted later.'''

    print "Starting calculation on interval {} to {}".format(times[0],times[1])

    osfilehandle,outfilename = tempfile.mkstemp(suffix=".fits")
    filter['rad'] = "INDEF"
    filter['evclass'] = "INDEF"
    filter['evclsmin'] = 0
    filter['evclsmax'] = 10
    filter['infile'] = times[3]
    filter['outfile'] = outfilename
    filter['ra'] = "INDEF"
    filter['dec'] = "INDEF"
    filter['tmin'] = times[0]
    filter['tmax'] = times[1]
    filter['emin'] = 0
    filter['emax'] = 400000
    filter['zmax'] = 180
    filter['convtype'] = -1
    filter['chatter'] = 0
    filter.run(print_command=False)

    diffResps['evfile'] = outfilename
    diffResps['scfile'] =  times[2]
    diffResps['srcmdl'] = times[4]
    diffResps['irfs'] = times[5]
    diffResps['chatter'] = 0
    diffResps.run(print_command=False)
    print "Completed calculation of interval {} to {}".format(times[0],times[1])
    return outfilename

def eventsum(filenames, Outfile, SaveTemp):

    '''This function takes a list of event files and sums them up using
    gtselect.  It first checks to see if there's only one temporary
    file.  If so, it just copies that to the output file.  If not, it
    creates a temporary file that lists the individual event files
    and operates gtselect on them.'''

    if len(filenames) <= 1:
        subprocess.call(["cp", filenames[0], Outfile])
    else:
        fileListfile = tempfile.NamedTemporaryFile()
        for filename in filenames:
            fileListfile.file.write(filename + "\n")
        fileListfile.flush()
        filter['rad'] = "INDEF"
        filter['evclass'] = "INDEF"
        filter['evclsmin'] = 0
        filter['evclsmax'] = 10
        filter['infile'] = "@"+fileListfile.name
        filter['outfile'] = Outfile
        filter['ra'] = "INDEF"
        filter['dec'] = "INDEF"
        filter['tmin'] = "INDEF"
        filter['tmax'] = "INDEF"
        filter['emin'] = 0
        filter['emax'] = 400000
        filter['zmax'] = 180
        filter['convtype'] = -1
        filter['chatter'] = 0
        filter.run(print_command=False)

    if SaveTemp:
        print "Did not delete the following temporary files:"
        print filenames
    else:
        print "Deleting temporary files..."
        for filename in filenames:
            os.remove(filename)


def gtdiffrsp_mp(bins, SCFile, EVFile, OutFile, SaveTemp, SrcModel,IRF):

    '''This function looks at the start and stop times in an event file
    and splits the time into chunks.  It then submits jobs based upon
    those start and stop times.'''

    print "Opening event file to determine break points..."
    hdulist = pyfits.open(EVFile)
    tstart = hdulist[0].header['TSTART']
    tstop = hdulist[0].header['TSTOP']
    hdulist.close()
    starts, step = np.linspace(tstart,tstop,bins,endpoint=False, retstep=True)
    stops = starts + step
    scfiles = [SCFile for st in starts]
    evfiles = [EVFile for st in starts]
    srcmdls = [SrcModel for st in starts]
    irfs =  [IRF for st in starts]

    pool = Pool(processes=bins)      
    times = np.array([starts,stops,scfiles,evfiles,srcmdls,irfs])

    print "Spawning {} jobs...".format(bins)
    tempfilenames = pool.map(diffrsp,times.transpose())
    print "Combining temporary files..."
    eventsum(tempfilenames, OutFile, SaveTemp)

def cli():

    helpString = "Submits the gtdiffrsp program as sperate threads via python and\
                  joins up the resulting temporary files at the end resulting in \
                  a single file.  This greatly reduces the running time. For more\
                  details on gtdiffrsp see the gtdiffrsp help file."

    import argparse

    parser = argparse.ArgumentParser(description=helpString)
    parser.add_argument("jobs", type=int, help="The number of jobs you wish to spawn (usually the number of cores on your machine).")
    
    parser.add_argument("evfile", help="Input event file.  See gtdiffrsp help for more information.")
    parser.add_argument("scfile", help="Spacecraft file.  See gtdiffrsp help for more information.")
    parser.add_argument("srcmdl", help="The source model file to use. See gtdiffrsp help for more information.")    
    parser.add_argument("irf", help="The IRFs to use. See gtdiffrsp help for more information.")    
    parser.add_argument("outfile", help="Output file name.")

    parser.add_argument("--savetmp", default = False, help="Save the temporary files (default is False).")

    args = parser.parse_args()

    gtdiffrsp_mp(args.jobs, args.scfile, args.evfile, args.outfile, args.savetmp, args.srcmdl, args.irf)

if __name__ == '__main__': cli()

    
